@@ -13,7 +13,7 @@
 ##############################################################################
 """Database connection support
 
-$Id: Connection.py 121409 2011-04-12 13:50:07Z jim $"""
+$Id$"""
 
 import logging
 import sys
@@ -321,10 +321,10 @@ class Connection(ExportImport, object):
                 # get back here.
         else:
             self.opened = None
-            
+
         am = self._db._activity_monitor
         if am is not None:
-            am.closedConnection(self)        
+            am.closedConnection(self)
 
     def db(self):
         """Returns a handle to the database this connection belongs to."""
@@ -995,7 +995,8 @@ class Connection(ExportImport, object):
     def readCurrent(self, ob):
         assert ob._p_jar is self
         assert ob._p_oid is not None and ob._p_serial is not None
-        self._readCurrent[ob._p_oid] = ob._p_serial
+        if ob._p_serial != z64:
+            self._readCurrent[ob._p_oid] = ob._p_serial
 
     # persistent.interfaces.IPersistentDatamanager
     ##########################################################################
@@ -1072,7 +1073,8 @@ class Connection(ExportImport, object):
     def _release_resources(self):
         for c in self.connections.itervalues():
             if c._mvcc_storage:
-                c._storage.release()
+                if c._storage is not None:
+                    c._storage.release()
             c._storage = c._normal_storage = None
             c._cache = PickleCache(self, 0, 0)
 
