@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
+# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -14,10 +14,11 @@
 """Storage implementation using a log written to a single file.
 """
 
-from cPickle import Pickler, Unpickler, loads
+from __future__ import with_statement
+
+from cPickle import Pickler, loads
 from persistent.TimeStamp import TimeStamp
 from struct import pack, unpack
-from types import StringType
 from zc.lockfile import LockFile
 from ZODB.FileStorage.format import CorruptedError, CorruptedDataError
 from ZODB.FileStorage.format import FileStorageFormatter, DataHeader
@@ -26,16 +27,15 @@ from ZODB.FileStorage.format import TxnHeader, DATA_HDR, DATA_HDR_LEN
 from ZODB.FileStorage.fspack import FileStoragePacker
 from ZODB.fsIndex import fsIndex
 from ZODB import BaseStorage, ConflictResolution, POSException
-from ZODB.loglevels import BLATHER
 from ZODB.POSException import UndoError, POSKeyError, MultipleUndoErrors
 from ZODB.utils import p64, u64, z64
 
 import base64
-import BTrees.OOBTree
+import contextlib
 import errno
 import logging
 import os
-import sys
+import threading
 import time
 import ZODB.blob
 import ZODB.interfaces
@@ -51,7 +51,7 @@ logger = logging.getLogger('ZODB.FileStorage')
 
 def panic(message, *data):
     logger.critical(message, *data)
-    raise CorruptedTransactionError(message)
+    raise CorruptedTransactionError(message % data)
 
 class FileStorageError(POSException.StorageError):
     pass
@@ -128,7 +128,7 @@ class FileStorage(
         else:
             self._tfile = None
 
-        self._file_name = file_name
+        self._file_name = os.path.abspath(file_name)
 
         self._pack_gc = pack_gc
         self.pack_keep_old = pack_keep_old
@@ -167,6 +167,7 @@ class FileStorage(
             self._file = open(file_name, 'w+b')
             self._file.write(packed_version)
 
+        self._files = FilePool(self._file_name)
         r = self._restore_index()
         if r is not None:
             self._used_index = 1 # Marker for testing
@@ -244,23 +245,7 @@ class FileStorage(
         index_name = self.__name__ + '.index'
         tmp_name = index_name + '.index_tmp'
 
-        f=open(tmp_name,'wb')
-        p=Pickler(f,1)
-
-        # Pickle the index buckets first to avoid deep recursion:
-        buckets = []
-        bucket = self._index._data._firstbucket
-        while bucket is not None:
-            buckets.append(bucket)
-            bucket = bucket._next
-        buckets.reverse()
-
-        info=BTrees.OOBTree.Bucket(dict(
-            _buckets=buckets, index=self._index, pos=self._pos))
-
-        p.dump(info)
-        f.flush()
-        f.close()
+        self._index.save(self._pos, tmp_name)
 
         try:
             try:
@@ -355,19 +340,15 @@ class FileStorage(
         file_name=self.__name__
         index_name=file_name+'.index'
 
-        try:
-            f = open(index_name, 'rb')
-        except:
+        if os.path.exists(index_name):
+            try:
+                info = fsIndex.load(index_name)
+            except:
+                logger.exception('loading index')
+                return None
+        else:
             return None
 
-        p=Unpickler(f)
-
-        try:
-            info=p.load()
-        except:
-            exc, err = sys.exc_info()[:2]
-            logger.warning("Failed to load database index: %s: %s", exc, err)
-            return None
         index = info.get('index')
         pos = info.get('pos')
         if index is None or pos is None:
@@ -401,6 +382,7 @@ class FileStorage(
 
     def close(self):
         self._file.close()
+        self._files.close()
         if hasattr(self,'_lock_file'):
             self._lock_file.close()
         if self._tfile:
@@ -426,48 +408,41 @@ class FileStorage(
         """Return pickle data and serial number."""
         assert not version
 
-        self._lock_acquire()
-        try:
+        with self._files.get() as _file:
             pos = self._lookup_pos(oid)
-            h = self._read_data_header(pos, oid)
+            h = self._read_data_header(pos, oid, _file)
             if h.plen:
-                data = self._file.read(h.plen)
+                data = _file.read(h.plen)
                 return data, h.tid
             elif h.back:
                 # Get the data from the backpointer, but tid from
                 # current txn.
-                data = self._loadBack_impl(oid, h.back)[0]
+                data = self._loadBack_impl(oid, h.back, _file=_file)[0]
                 return data, h.tid
             else:
                 raise POSKeyError(oid)
-        finally:
-            self._lock_release()
 
     def loadSerial(self, oid, serial):
-        self._lock_acquire()
-        try:
+        with self._lock:
             pos = self._lookup_pos(oid)
             while 1:
                 h = self._read_data_header(pos, oid)
                 if h.tid == serial:
                     break
                 pos = h.prev
-                if not pos:
+                if h.tid < serial or not pos:
                     raise POSKeyError(oid)
             if h.plen:
                 return self._file.read(h.plen)
             else:
                 return self._loadBack_impl(oid, h.back)[0]
-        finally:
-            self._lock_release()
 
     def loadBefore(self, oid, tid):
-        self._lock_acquire()
-        try:
+        with self._files.get() as _file:
             pos = self._lookup_pos(oid)
             end_tid = None
             while True:
-                h = self._read_data_header(pos, oid)
+                h = self._read_data_header(pos, oid, _file)
                 if h.tid < tid:
                     break
 
@@ -477,13 +452,10 @@ class FileStorage(
                     return None
 
             if h.back:
-                data, _, _, _ = self._loadBack_impl(oid, h.back)
+                data, _, _, _ = self._loadBack_impl(oid, h.back, _file=_file)
                 return data, h.tid, end_tid
             else:
-                return self._file.read(h.plen), h.tid, end_tid
-
-        finally:
-            self._lock_release()
+                return _file.read(h.plen), h.tid, end_tid
 
     def store(self, oid, oldserial, data, version, transaction):
         if self._is_read_only:
@@ -492,8 +464,7 @@ class FileStorage(
             raise POSException.StorageTransactionError(self, transaction)
         assert not version
 
-        self._lock_acquire()
-        try:
+        with self._lock:
             if oid > self._oid:
                 self.set_max_oid(oid)
             old = self._index_get(oid, 0)
@@ -504,14 +475,8 @@ class FileStorage(
                 committed_tid = h.tid
 
                 if oldserial != committed_tid:
-                    rdata = self.tryToResolveConflict(oid, committed_tid,
+                    data = self.tryToResolveConflict(oid, committed_tid,
                                                      oldserial, data)
-                    if rdata is None:
-                        raise POSException.ConflictError(
-                            oid=oid, serials=(committed_tid, oldserial),
-                            data=data)
-                    else:
-                        data = rdata
 
             pos = self._pos
             here = pos + self._tfile.tell() + self._thl
@@ -531,17 +496,13 @@ class FileStorage(
             else:
                 return self._tid
 
-        finally:
-            self._lock_release()
-
     def deleteObject(self, oid, oldserial, transaction):
         if self._is_read_only:
             raise POSException.ReadOnlyError()
         if transaction is not self._transaction:
             raise POSException.StorageTransactionError(self, transaction)
 
-        self._lock_acquire()
-        try:
+        with self._lock:
             old = self._index_get(oid, 0)
             if not old:
                 raise POSException.POSKeyError(oid)
@@ -564,9 +525,6 @@ class FileStorage(
                 raise FileStorageQuotaError(
                     "The storage quota has been exceeded.")
 
-        finally:
-            self._lock_release()
-
     def _data_find(self, tpos, oid, data):
         # Return backpointer for oid.  Must call with the lock held.
         # This is a file offset to oid's data record if found, else 0.
@@ -638,8 +596,7 @@ class FileStorage(
         if version:
             raise TypeError("Versions are no-longer supported")
 
-        self._lock_acquire()
-        try:
+        with self._lock:
             if oid > self._oid:
                 self.set_max_oid(oid)
             prev_pos = 0
@@ -675,8 +632,6 @@ class FileStorage(
                     self._tfile.write(z64)
             else:
                 self._tfile.write(data)
-        finally:
-            self._lock_release()
 
     def supportsUndo(self):
         return 1
@@ -700,12 +655,11 @@ class FileStorage(
             if len(e) > 65535:
                 raise FileStorageError('too much extension data')
 
-
     def tpc_vote(self, transaction):
-        self._lock_acquire()
-        try:
+        with self._lock:
             if transaction is not self._transaction:
-                return
+                raise POSException.StorageTransactionError(
+                    "tpc_vote called with wrong transaction")
             dlen = self._tfile.tell()
             if not dlen:
                 return # No data in this trans
@@ -729,10 +683,26 @@ class FileStorage(
                 # Hm, an error occurred writing out the data. Maybe the
                 # disk is full. We don't want any turd at the end.
                 self._file.truncate(self._pos)
+                self._files.flush()
                 raise
             self._nextpos = self._pos + (tl + 8)
-        finally:
-            self._lock_release()
+
+    def tpc_finish(self, transaction, f=None):
+        with self._files.write_lock():
+            with self._lock:
+                if transaction is not self._transaction:
+                    raise POSException.StorageTransactionError(
+                        "tpc_finish called with wrong transaction")
+                try:
+                    if f is not None:
+                        f(self._tid)
+                    u, d, e = self._ude
+                    self._finish(self._tid, u, d, e)
+                    self._clear_temp()
+                finally:
+                    self._ude = None
+                    self._transaction = None
+                    self._commit_lock_release()
 
     def _finish(self, tid, u, d, e):
         # If self._nextpos is 0, then the transaction didn't write any
@@ -768,6 +738,7 @@ class FileStorage(
     def _abort(self):
         if self._nextpos:
             self._file.truncate(self._pos)
+            self._files.flush()
             self._nextpos=0
             self._blob_tpc_abort()
 
@@ -797,16 +768,13 @@ class FileStorage(
         return h.tid, pos, data
 
     def getTid(self, oid):
-        self._lock_acquire()
-        try:
+        with self._lock:
             pos = self._lookup_pos(oid)
             h = self._read_data_header(pos, oid)
             if h.plen == 0 and h.back == 0:
                 # Undone creation
                 raise POSKeyError(oid)
             return h.tid
-        finally:
-            self._lock_release()
 
     def _transactionalUndoRecord(self, oid, pos, tid, pre):
         """Get the undo information for a data record
@@ -870,10 +838,11 @@ class FileStorage(
         except KeyError:
             # couldn't find oid; what's the real explanation for this?
             raise UndoError("_loadBack() failed for %s", oid)
-        data = self.tryToResolveConflict(oid, ctid, tid, bdata, cdata)
-
-        if data:
+        try:
+            data = self.tryToResolveConflict(oid, ctid, tid, bdata, cdata)
             return data, 0, ipos
+        except POSException.ConflictError:
+            pass
 
         raise UndoError("Some data were modified by a later transaction", oid)
 
@@ -890,8 +859,7 @@ class FileStorage(
             # the normalization code was incorrect for years (used +1
             # instead -- off by 1), until ZODB 3.4.
             last = first - last
-        self._lock_acquire()
-        try:
+        with self._lock:
             if self._pack_is_in_progress:
                 raise UndoError(
                     'Undo is currently disabled for database maintenance.<p>')
@@ -908,8 +876,6 @@ class FileStorage(
                 self._lock_release()
                 self._lock_acquire()
             return us.results
-        finally:
-            self._lock_release()
 
     def undo(self, transaction_id, transaction):
         """Undo a transaction, given by transaction_id.
@@ -928,20 +894,14 @@ class FileStorage(
         if transaction is not self._transaction:
             raise POSException.StorageTransactionError(self, transaction)
 
-        self._lock_acquire()
-        try:
-            return self._txn_undo(transaction_id)
-        finally:
-            self._lock_release()
-
-    def _txn_undo(self, transaction_id):
-        # Find the right transaction to undo and call _txn_undo_write().
-        tid = base64.decodestring(transaction_id + '\n')
-        assert len(tid) == 8
-        tpos = self._txn_find(tid, 1)
-        tindex = self._txn_undo_write(tpos)
-        self._tindex.update(tindex)
-        return self._tid, tindex.keys()
+        with self._lock:
+          # Find the right transaction to undo and call _txn_undo_write().
+          tid = base64.decodestring(transaction_id + '\n')
+          assert len(tid) == 8
+          tpos = self._txn_find(tid, 1)
+          tindex = self._txn_undo_write(tpos)
+          self._tindex.update(tindex)
+          return self._tid, tindex.keys()
 
     def _txn_find(self, tid, stop_at_pack):
         pos = self._pos
@@ -997,7 +957,7 @@ class FileStorage(
                     except ZODB.POSException.POSKeyError:
                         pass # It was removed, so no need to copy data
                     else:
-                        if ZODB.blob.is_blob_record(up):
+                        if self.is_blob_record(up):
                             # We're undoing a blob modification operation.
                             # We have to copy the blob data
                             tmp = ZODB.utils.mktemp(dir=self.fshelper.temp_dir)
@@ -1030,8 +990,7 @@ class FileStorage(
         return tindex
 
     def history(self, oid, size=1, filter=None):
-        self._lock_acquire()
-        try:
+        with self._lock:
             r = []
             pos = self._lookup_pos(oid)
 
@@ -1059,8 +1018,6 @@ class FileStorage(
                     pos = h.prev
                 else:
                     return r
-        finally:
-            self._lock_release()
 
     def _redundant_pack(self, file, pos):
         assert pos > 8, pos
@@ -1100,13 +1057,10 @@ class FileStorage(
         if not self._index:
             return
 
-        self._lock_acquire()
-        try:
+        with self._lock:
             if self._pack_is_in_progress:
                 raise FileStorageError('Already packing')
             self._pack_is_in_progress = True
-        finally:
-            self._lock_release()
 
         if gc is None:
             gc = self._pack_gc
@@ -1130,22 +1084,21 @@ class FileStorage(
                 return
             have_commit_lock = True
             opos, index = pack_result
-            self._lock_acquire()
-            try:
-                self._file.close()
-                try:
-                    os.rename(self._file_name, oldpath)
-                except Exception:
-                    self._file = open(self._file_name, 'r+b')
-                    raise
+            with self._files.write_lock():
+                with self._lock:
+                    self._files.empty()
+                    self._file.close()
+                    try:
+                        os.rename(self._file_name, oldpath)
+                    except Exception:
+                        self._file = open(self._file_name, 'r+b')
+                        raise
 
-                # OK, we're beyond the point of no return
-                os.rename(self._file_name + '.pack', self._file_name)
-                self._file = open(self._file_name, 'r+b')
-                self._initIndex(index, self._tindex)
-                self._pos = opos
-            finally:
-                self._lock_release()
+                    # OK, we're beyond the point of no return
+                    os.rename(self._file_name + '.pack', self._file_name)
+                    self._file = open(self._file_name, 'r+b')
+                    self._initIndex(index, self._tindex)
+                    self._pos = opos
 
             # We're basically done.  Now we need to deal with removed
             # blobs and removing the .old file (see further down).
@@ -1158,18 +1111,14 @@ class FileStorage(
         finally:
             if have_commit_lock:
                 self._commit_lock_release()
-            self._lock_acquire()
-            self._pack_is_in_progress = False
-            self._lock_release()
+            with self._lock:
+                self._pack_is_in_progress = False
 
         if not self.pack_keep_old:
             os.remove(oldpath)
 
-        self._lock_acquire()
-        try:
+        with self._lock:
             self._save_index()
-        finally:
-            self._lock_release()
 
     def _remove_blob_files_tagged_for_removal_during_pack(self):
         lblob_dir = len(self.blob_dir)
@@ -1273,16 +1222,11 @@ class FileStorage(
     def iterator(self, start=None, stop=None):
         return FileIterator(self._file_name, start, stop)
 
-    def lastTransaction(self):
-        """Return transaction id for last committed transaction"""
-        return self._ltid
-
     def lastInvalidations(self, count):
         file = self._file
         seek = file.seek
         read = file.read
-        self._lock_acquire()
-        try:
+        with self._lock:
             pos = self._pos
             while count > 0 and pos > 4:
                 count -= 1
@@ -1292,9 +1236,6 @@ class FileStorage(
             seek(0)
             return [(trans.tid, [r.oid for r in trans])
                     for trans in FileIterator(self._file_name, pos=pos)]
-        finally:
-            self._lock_release()
-
 
     def lastTid(self, oid):
         """Return last serialno committed for object oid.
@@ -1331,7 +1272,15 @@ class FileStorage(
 
         return oid, tid, data, next_oid
 
+    ######################################################################
+    # The following 2 methods are for testing a ZEO extension mechanism
+    def getExtensionMethods(self):
+        return dict(answer_to_the_ultimate_question=None)
 
+    def answer_to_the_ultimate_question(self):
+        return 42
+    #
+    ######################################################################
 
 def shift_transactions_forward(index, tindex, file, pos, opos):
     """Copy transactions forward in the data file
@@ -1819,7 +1768,7 @@ class FileIterator(FileStorageFormatter):
 
     def next(self):
         if self._file is None:
-            raise ZODB.interfaces.StorageStopIteration()
+            raise StopIteration()
 
         pos = self._pos
         while True:
@@ -1908,7 +1857,7 @@ class FileIterator(FileStorageFormatter):
             return result
 
         self.close()
-        raise ZODB.interfaces.StorageStopIteration()
+        raise StopIteration()
 
 
 class TransactionRecord(BaseStorage.TransactionRecord):
@@ -1967,7 +1916,7 @@ class TransactionRecordIterator(FileStorageFormatter):
 
             return Record(h.oid, h.tid, data, prev_txn, pos)
 
-        raise ZODB.interfaces.StorageStopIteration()
+        raise StopIteration()
 
 
 class Record(BaseStorage.DataRecord):
@@ -2036,3 +1985,81 @@ class UndoSearch:
              'description': d}
         d.update(e)
         return d
+
+class FilePool:
+
+    closed = False
+    writing = False
+    writers = 0
+
+    def __init__(self, file_name):
+        self.name = file_name
+        self._files = []
+        self._out = []
+        self._cond = threading.Condition()
+
+    @contextlib.contextmanager
+    def write_lock(self):
+        with self._cond:
+            self.writers += 1
+            while self.writing or self._out:
+                self._cond.wait()
+            if self.closed:
+                raise ValueError('closed')
+            self.writing = True
+
+        try:
+            yield None
+        finally:
+            with self._cond:
+                self.writing = False
+                if self.writers > 0:
+                    self.writers -= 1
+                self._cond.notifyAll()
+
+    @contextlib.contextmanager
+    def get(self):
+        with self._cond:
+            while self.writers:
+                self._cond.wait()
+            assert not self.writing
+            if self.closed:
+                raise ValueError('closed')
+
+            try:
+                f = self._files.pop()
+            except IndexError:
+                f = open(self.name, 'rb')
+            self._out.append(f)
+
+        try:
+            yield f
+        finally:
+            self._out.remove(f)
+            self._files.append(f)
+            if not self._out:
+                with self._cond:
+                    if self.writers and not self._out:
+                        self._cond.notifyAll()
+
+    def empty(self):
+        while self._files:
+            self._files.pop().close()
+
+
+    def flush(self):
+        """Empty read buffers.
+
+        This is required if they contain data of rolled back transactions.
+        """
+        with self.write_lock():
+            for f in self._files:
+                f.flush()
+
+    def close(self):
+        with self._cond:
+            self.closed = True
+            while self._out:
+                self._out.pop().close()
+            self.empty()
+            self.writing = self.writers = 0
