@@ -13,12 +13,16 @@
 ##############################################################################
 """Broken object support
 
-$Id: broken.py 113733 2010-06-21 15:32:58Z ctheune $
+$Id$
 """
 
 import sys
 import persistent
 
+import zope.interface
+
+import ZODB.interfaces
+
 broken_cache = {}
 
 class Broken(object):
@@ -92,6 +96,8 @@ class Broken(object):
          >>> broken_cache.clear()
        """
 
+    zope.interface.implements(ZODB.interfaces.IBroken)
+
     __Broken_state__ = __Broken_initargs__ = None
 
     __name__ = 'broken object'
@@ -177,6 +183,13 @@ def find_global(modulename, globalname,
 
          >>> broken_cache.clear()
        """
+
+    # short circuit common case:
+    try:
+        return getattr(sys.modules[modulename], globalname)
+    except (AttributeError, KeyError):
+        pass
+
     try:
         __import__(modulename)
     except ImportError:
