@@ -286,24 +286,41 @@ class IConnection(Interface):
         begins or until the connection os reopned.
         """
 
+    def readCurrent(obj):
+        """Make sure an object being read is current
 
-class IStorageDB(Interface):
-    """Database interface exposed to storages
+        This is used when applications want to ensure a higher level
+        of consistency for some operations. This should be called when
+        an object is read and the information read is used to write a
+        separate object.
+        """
 
-    This interface provides 2 facilities:
+class IStorageWrapper(Interface):
+    """Storage wrapper interface
+
+    This interface provides 3 facilities:
 
     - Out-of-band invalidation support
 
-      A storage can notify it's database of object invalidations that
+      A storage can notify it's wrapper of object invalidations that
       don't occur due to direct operations on the storage.  Currently
       this is only used by ZEO client storages to pass invalidation
       messages sent from a server.
 
-    - Record-reference extraction.
+    - Record-reference extraction
 
       The references method can be used to extract referenced object
       IDs from a database record.  This can be used by storages to
-      provide more advanced garbage collection.
+      provide more advanced garbage collection.  A wrapper storage
+      that transforms data will provide a references method that
+      untransforms data passed to it and then pass the data to the
+      layer above it.
+
+    - Record transformation
+
+      A storage wrapper may transform data, for example for
+      compression or encryption.  Methods are provided to transform or
+      untransform data.
 
     This interface may be implemented by storage adapters or other
     intermediaries.  For example, a storage adapter that provides
@@ -316,7 +333,7 @@ class IStorageDB(Interface):
 
         This can be necessary if there have been major changes to
         stored data and it is either impractical to enumerate them or
-        there would be so many that it would be inefficient to do so.        
+        there would be so many that it would be inefficient to do so.
         """
 
     def invalidate(transaction_id, oids, version=''):
@@ -326,7 +343,7 @@ class IStorageDB(Interface):
 
         The version argument is provided for backward
         compatibility. If passed, it must be an empty string.
-        
+
         """
 
     def references(record, oids=None):
@@ -337,13 +354,23 @@ class IStorageDB(Interface):
         be created and returned.
         """
 
+    def transform_record_data(data):
+        """Return transformed data
+        """
+
+    def untransform_record_data(data):
+        """Return untransformed data
+        """
+
+IStorageDB = IStorageWrapper # for backward compatibility
+
 
 class IDatabase(IStorageDB):
     """ZODB DB.
     """
 
     # TODO: This interface is incomplete.
-    # XXX how is it incomplete? 
+    # XXX how is it incomplete?
 
     databases = Attribute(
         """A mapping from database name to DB (database) object.
@@ -362,7 +389,7 @@ class IDatabase(IStorageDB):
         application code should rarely, if ever, have a need to use
         this attribute.
         """)
-    
+
 
     def open(transaction_manager=None, serial=''):
         """Return an IConnection object for use by application code.
@@ -431,6 +458,36 @@ class IDatabase(IStorageDB):
 
 class IStorage(Interface):
     """A storage is responsible for storing and retrieving data of objects.
+
+    Consistency and locking
+    -----------------------
+
+    When transactions are committed, a storage assigns monotonically
+    increasing transaction identifiers (tids) to the transactions and
+    to the object versions written by the transactions.  ZODB relies
+    on this to decide if data in object caches are up to date and to
+    implement multi-version concurrency control.
+
+    There are methods in IStorage and in derived interfaces that
+    provide information about the current revisions (tids) for objects
+    or for the database as a whole.  It is critical for the proper
+    working of ZODB that the resulting tids are increasing with
+    respect to the object identifier given or to the databases.  That
+    is, if there are 2 results for an object or for the database, R1
+    and R2, such that R1 is returned before R2, then the tid returned
+    by R2 must be greater than or equal to the tid returned by R1.
+    (When thinking about results for the database, think of these as
+    results for all objects in the database.)
+
+    This implies some sort of locking strategy.  The key method is
+    tcp_finish, which causes new tids to be generated and also,
+    through the callback passed to it, returns new current tids for
+    the objects stored in a transaction and for the database as a whole.
+
+    The IStorage methods affected are lastTransaction, load, store,
+    and tpc_finish.  Derived interfaces may introduce additional
+    methods.
+
     """
 
     def close():
@@ -451,7 +508,7 @@ class IStorage(Interface):
 
     def getSize():
         """An approximate size of the database, in bytes.
-        
+
         This is used soley for informational purposes.
         """
 
@@ -459,7 +516,7 @@ class IStorage(Interface):
         """Return a sequence of history information dictionaries.
 
         Up to size objects (including no objects) may be returned.
-        
+
         The information provides a log of the changes made to the
         object. Data are reported in reverse chronological order.
 
@@ -468,7 +525,7 @@ class IStorage(Interface):
         time
             UTC seconds since the epoch (as in time.time) that the
             object revision was committed.
-            
+
         tid
             The transaction identifier of the transaction that
             committed the version.
@@ -489,7 +546,7 @@ class IStorage(Interface):
 
         If the transaction had extension items, then these items are
         also included if they don't conflict with the keys above.
-        
+
         """
 
     def isReadOnly():
@@ -505,12 +562,15 @@ class IStorage(Interface):
         # transiently.  It would be better to just have read-only errors.
 
     def lastTransaction():
-        """Return the id of the last committed transaction
+        """Return the id of the last committed transaction.
+
+        If no transactions have been committed, return a string of 8
+        null (0) characters.
         """
 
     def __len__():
         """The approximate number of objects in the storage
-        
+
         This is used soley for informational purposes.
         """
 
@@ -595,12 +655,18 @@ class IStorage(Interface):
         revisions.
         """
 
-    def registerDB(db):
-        """Register an IStorageDB.
+    def registerDB(wrapper):
+        """Register a storage wrapper IStorageWrapper.
+
+        The passed object is a wrapper object that provides an upcall
+        interface to support composition.
 
         Note that, for historical reasons, an implementation may
         require a second argument, however, if required, the None will
         be passed as the second argument.
+
+        Also, for historical reasons, this is called registerDB rather
+        than register_wrapper.
         """
 
     def sortKey():
@@ -623,7 +689,7 @@ class IStorage(Interface):
         oid
             The object identifier.  This is either a string
             consisting of 8 nulls or a string previously returned by
-            new_oid. 
+            new_oid.
 
         serial
             The serial of the data that was read when the object was
@@ -647,15 +713,14 @@ class IStorage(Interface):
 
         The return value may be:
 
-        - None
+        - None, or
 
-        - A new serial (string) for the object, or
+        - A new serial (string) for the object
 
-        - An iterable of object-id and serial pairs giving new serials
-          for objects.
+        If None is returned, then a new serial (or other special
+        values) must ve returned in tpc_vote results.
 
-        A serial, returned as a string or in a sequence of oid/serial
-        pairs, may be the special value
+        A serial, returned as a string, may be the special value
         ZODB.ConflictResolution.ResolvedSerial to indicate that a
         conflict occured and that the object should be invalidated.
 
@@ -673,7 +738,7 @@ class IStorage(Interface):
         StorageError or, more often, a subclass of it
           is raised when an internal error occurs while the storage is
           handling the store() call.
-        
+
         """
 
     def tpc_abort(transaction):
@@ -690,7 +755,7 @@ class IStorage(Interface):
         """Begin the two-phase commit process.
 
         If storage is already participating in a two-phase commit
-        using the same transaction, the call is ignored.
+        using the same transaction, a StorageTransactionError is raised.
 
         If the storage is already participating in a two-phase commit
         using a different transaction, the call blocks until the
@@ -702,9 +767,10 @@ class IStorage(Interface):
 
         Changes must be made permanent at this point.
 
-        This call is ignored if the storage isn't participating in
-        two-phase commit or if it is committing a different
-        transaction.  Failure of this method is extremely serious.
+        This call raises a StorageTransactionError if the storage
+        isn't participating in two-phase commit or if it is committing
+        a different transaction.  Failure of this method is extremely
+        serious.
 
         The second argument is a call-back function that must be
         called while the storage transaction lock is held.  It takes
@@ -715,9 +781,9 @@ class IStorage(Interface):
     def tpc_vote(transaction):
         """Provide a storage with an opportunity to veto a transaction
 
-        This call is ignored if the storage isn't participating in
-        two-phase commit or if it is commiting a different
-        transaction.  Failure of this method is extremely serious.
+        This call raises a StorageTransactionError if the storage
+        isn't participating in two-phase commit or if it is commiting
+        a different transaction.
 
         If a transaction can be committed by a storage, then the
         method should return.  If a transaction cannot be committed,
@@ -771,7 +837,7 @@ class IStorageRestoreable(IStorage):
         # - Incorrect pack garbage-collection algorithms (possibly
         #   including the existing FileStorage implementation), that
         #   failed to take into account records after the pack time.
-        
+
 
     def restore(oid, serial, data, version, prev_txn, transaction):
         """Write data already committed in a separate database
@@ -785,7 +851,7 @@ class IStorageRestoreable(IStorage):
 
         oid
              The object id for the record
-        
+
         serial
              The transaction identifier that originally committed this object.
 
@@ -1071,7 +1137,7 @@ class IStorageCurrentRecordIteration(IStorage):
             ...     # do things with oid, tid, and data
             ...     if next is None:
             ...         break
-        
+
         """
 
 class IExternalGC(IStorage):
@@ -1093,6 +1159,22 @@ class IExternalGC(IStorage):
        commit.
        """
 
+class ReadVerifyingStorage(IStorage):
+
+    def checkCurrentSerialInTransaction(oid, serial, transaction):
+        """Check whether the given serial number is current.
+
+        The method is called during the first phase of 2-phase commit
+        to verify that data read in a transaction is current.
+
+        The storage should raise a ReadConflictError if the serial is not
+        current, although it may raise the exception later, in a call
+        to store or in a call to tpc_vote.
+
+        If no exception is raised, then the serial must remain current
+        through the end of the transaction.
+        """
+
 class IBlob(Interface):
     """A BLOB supports efficient handling of large data within ZODB."""
 
@@ -1106,7 +1188,7 @@ class IBlob(Interface):
         The mode 'c' is similar to 'r', except that an orinary file
         object is returned and may be used in a separate transaction
         and after the blob's database connection has been closed.
-        
+
         """
 
     def committed():
@@ -1127,12 +1209,12 @@ class IBlob(Interface):
         Replace the current data of the blob with the file given under
         filename.
 
-        The blob must not be opened for reading or writing when consuming a 
+        The blob must not be opened for reading or writing when consuming a
         file.
 
         The blob will take over ownership of the file and will either
         rename or copy and remove it.  The file must not be open.
-        
+
         """
 
 
@@ -1147,13 +1229,45 @@ class IBlobStorage(Interface):
         (or copy and remove it) immediately, or at transaction-commit
         time.  The file must not be open.
 
-        The new serial is returned.
+        The new serial for the object is returned, but not necessarily
+        immediately.  It may be returned directly, or on a subsequent
+        store or tpc_vote call.
+
+        The return value may be:
+
+        - None
+
+        - A new serial (string) for the object, or
+
+        - An iterable of object-id and serial pairs giving new serials
+          for objects.
+
+        A serial, returned as a string or in a sequence of oid/serial
+        pairs, may be the special value
+        ZODB.ConflictResolution.ResolvedSerial to indicate that a
+        conflict occured and that the object should be invalidated.
+
+        Several different exceptions may be raised when an error occurs.
+
+        ConflictError
+          is raised when serial does not match the most recent serial
+          number for object oid and the conflict was not resolved by
+          the storage.
+
+        StorageTransactionError
+          is raised when transaction does not match the current
+          transaction.
+
+        StorageError or, more often, a subclass of it
+          is raised when an internal error occurs while the storage is
+          handling the store() call.
+
         """
 
     def loadBlob(oid, serial):
         """Return the filename of the Blob data for this OID and serial.
 
-        Returns a filename. 
+        Returns a filename.
 
         Raises POSKeyError if the blobfile cannot be found.
         """
@@ -1185,6 +1299,31 @@ class IBlobStorageRestoreable(IBlobStorage, IStorageRestoreable):
         """
 
 
+class IBroken(Interface):
+    """Broken objects are placeholders for objects that can no longer be
+    created because their class has gone away.
+
+    They cannot be modified, but they retain their state. This allows them to
+    be rebuild should the missing class be found again.
+
+    A broken object's __class__ can be used to determine the original
+    class' name (__name__) and module (__module__).
+
+    The original object's state and initialization arguments are
+    available in broken object attributes to aid analysis and
+    reconstruction.
+
+    """
+
+    def __setattr__(name, value):
+        """You cannot modify broken objects. This will raise a
+        ZODB.broken.BrokenModified exception.
+        """
+
+    __Broken_newargs__ = Attribute("Arguments passed to __new__.")
+    __Broken_initargs__ = Attribute("Arguments passed to __init__.")
+    __Broken_state__ = Attribute("Value passed to __setstate__.")
+
 class BlobError(Exception):
     pass
 
