@@ -4,21 +4,18 @@
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
-# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
+# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
 # THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
 # WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 # WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-"""Tests of savepoint feature
-
-$Id: testConnectionSavepoint.py 113733 2010-06-21 15:32:58Z ctheune $
-"""
-import unittest
-from zope.testing import doctest
+import doctest
 import persistent.mapping
 import transaction
+import unittest
+import ZODB.tests.util
 
 def testAddingThenModifyThenAbort():
     """\
@@ -182,6 +179,7 @@ the wrong state.
     1
     """
 
+
 def tearDown(test):
     transaction.abort()
 
